"""
Gradient checkpointing.

Following a lot of the code of the official
`torch.utils.checkpoint <https://pytorch.org/docs/stable/checkpoint.html>`__,
using ``torch.autograd.graph.saved_tensors_hooks``
and ``TorchDispatchMode``
but also handling the RNG fork and reset in a similar way.

See also :mod:`returnn.tf.util.gradient_checkpoint`:
same API and logic in TF, although it heavily makes use
of the TF computation graph, i.e. graph mode,
which makes this particular feature much easier to implement.

See also:
https://github.com/rwth-i6/returnn/issues/1552
https://discuss.pytorch.org/t/gradient-checkpointing/205416
https://gist.github.com/soulitzer/ec1049a947be046de7fbc2af61a4ee8c
"""

from __future__ import annotations

from typing import Optional, Union, Any, Callable, Sequence, List, Dict
from types import MethodType
from dataclasses import dataclass, field
import contextlib
from weakref import ref, WeakSet
import threading
import atexit

import torch
from torch.utils.weak import WeakTensorKeyDictionary  # needs Torch >=2.0.0

# noinspection PyProtectedMember
from torch.utils._python_dispatch import TorchDispatchMode

# PyTree is very common and semi-standard for PyTorch, e.g. __torch_dispatch__.
# We might use dm-tree or so alternatively here, but PyTree should be fine.
# noinspection PyProtectedMember
import torch.utils._pytree as pytree


__all__ = ["gradient_checkpoint_scope"]


# gradient_checkpoint_scope is the public API to the user.
# gradient_checkpoint_scope.__enter__ will enter two other scopes:
#
# - record_graph_scope: _RecordGraph(TorchDispatchMode),
#   to record the computation graph for all ops within the scope.
#
# - saved_tensors_hooks_scope: torch.autograd.graph.saved_tensors_hooks,
#   to overwrite what we store for backpropagation, and how to recompute it.
#   Specifically, for all tensors which were created within the gradient_checkpoint_scope,
#   we will never store them in the pack_hook,
#   and unpack_hook will trigger the recomputation of the computation graph.
#
# gradient_checkpoint_scope.__exit__ will exit the record_graph_scope,
# but the saved_tensors_hooks_scope will stay alive as long as needed,
# while any of the created tensors are still alive.
# We keep a weak tensor key dictionary to map from the created raw tensors
# to the point in the recorded computation graph (specifically _GraphTensor objects).
# We just check whether any of the weak tensor refs is still alive.
#
# To keep saved_tensors_hooks_scope alive and make sure
# that other calls to torch.autograd.graph.saved_tensors_hooks are correctly handled,
# specifically that the order of enter/exit is correct,
# we hook into torch.autograd.graph.saved_tensors_hooks.__enter__/__exit__ itself.
# See _register_custom_saved_tensors_hooks below.
# Further, torch.autograd.graph.saved_tensors_hooks is thread local,
# so we can do any such logic only within the same thread.
# We also hook into Tensor.__del__ and also handle gradient_checkpoint_scope.__del__,
# but as that might run in a different thread, we cannot always do the cleanup there.
# We always check for this.
# (Note that this is due to the API of torch.autograd.graph.saved_tensors_hooks.
# We actually would want to always use it for a set of specified tensors.
# We also discuss some potentially better PyTorch API to implement this in an easier way:
# https://github.com/pytorch/pytorch/issues/129867)
#
# For the recomputation, we make sure that we properly reset the RNG and AMP states,
# and that we perform the recomputation in the exact same order, such that RNG state is correct.
#
# Once some recomputed tensor was used and is not needed anymore, the GC should free it.
# We try to make sure that no unnecessary references are kept alive.
#
# Also see test_gradient_checkpoint_scope() which tests this.


class gradient_checkpoint_scope:
    """
    Create a gradient checkpoint scope.
    All tensors created within this scope will not be stored for backpropagation,
    but will be recomputed on the fly during backpropagation.

    Example::

        a = ...
        b = ...
        c = ...
        with gradient_checkpoint_scope():
            x = a + b
        y = x * c

    In this example, the tensor ``x`` will not be stored for backpropagation,
    i.e. the computation ``x = a + b`` will be recomputed during backpropagation.

    Internally, this uses the PyTorch ``torch.autograd.graph.saved_tensors_hooks`` mechanism
    to override what we store for backpropagation, and how to recompute it.
    And we use the PyTorch ``TorchDispatchMode`` to intercept all operations within the scope.
    Note that the usage of ``torch.autograd.graph.saved_tensors_hooks`` is tricky here
    as we need it beyond the scope of the ``gradient_checkpoint_scope``,
    specifically for all future usages of the tensor ``x`` in the example.
    See the code documentation for more details on this.

    Note, PyTorch itself also provides a gradient checkpointing API,
    namely `torch.utils.checkpoint <https://pytorch.org/docs/stable/checkpoint.html>`__.
    This API is different: You cannot easily specify what not to store / what to recompute.
    You rather specify a start/end point what to *store* for backpropagation,
    and then PyTorch will recompute everything in between.
    For the example above, you define that ``y`` is the end point and will be stored.
    It looks like this::

        a = ...
        b = ...
        c = ...
        y = torch.utils.checkpoint.checkpoint(lambda: (a + b) * c)

    PyTorch will not recompute ``... * c`` here,
    but it will recompute ``a + b``.
    We find this API more cumbersome to use and less flexible,
    because in many case, you know what you want to recompute, i.e. what you don't want to store.
    The PyTorch API is more about what you want to store, and then recompute everything else between.

    See also:
    https://github.com/rwth-i6/returnn/issues/1552
    https://discuss.pytorch.org/t/gradient-checkpointing/205416
    """

    def __init__(self):
        self.record_graph_scope = _RecordGraph()
        self.record_graph_scope.graph.gradient_checkpoint_scope_backref = self
        # Note: saved_tensors_hooks is thread local.
        self.saved_tensors_hooks_scope = torch.autograd.graph.saved_tensors_hooks(self._pack_hook, self._unpack_hook)
        self.entered = False
        self.entered_thread_ref = None
        self.exit_args: Optional[tuple] = None
        self.exited_saved_tensors_hooks_scope = False

    def __enter__(self):
        self.record_graph_scope.__enter__()
        self.saved_tensors_hooks_scope.__enter__()
        self.entered = True
        self.entered_thread_ref = ref(threading.current_thread())

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exit_args = (exc_type, exc_val, exc_tb)
        self.record_graph_scope.__exit__(exc_type, exc_val, exc_tb)
        if self.record_graph_scope.graph.is_any_recorded_tensor_alive():
            # Do not exit saved_tensors_hooks_scope here
            # because we still want to pack any tensors which were captured in our graph
            # by giving it a ref to the graph tensor.
            # However, we must track any further external calls to saved_tensors_hooks_scope,
            # to be able to properly remove it from the stack at the right point.
            _register_custom_saved_tensors_hooks(existing_scope=self.saved_tensors_hooks_scope)
            _register_custom_saved_tensors_hooks_thread_local_callback(
                _WeakMethod(self._custom_saved_tensors_hooks_callback, return_if_dead=False)
            )
        else:  # no relevant tensors alive anymore
            self.exit_saved_tensors_hooks_scope()

    def _maybe_exit_saved_tensors_hooks_scope(self):
        if self.exited_saved_tensors_hooks_scope:
            return
        if not self.exit_args:
            return
        # If we are in the right thread, maybe we can do the cleanup now.
        if self.entered_thread_ref() is threading.current_thread():
            if not self.record_graph_scope.graph.is_any_recorded_tensor_alive():
                self.exit_saved_tensors_hooks_scope()

    def __del__(self):
        if _python_exit:
            return
        # Note, be very careful what we do in __del__ because it might be called in a different thread!
        # Note that the __del__ will likely be called very late,
        # as the reference to the _Graph is kept alive until we used it for backprop,
        # as we keep this alive via _Graph.gradient_checkpoint_scope_backref
        # as long as any _GraphTensor is alive due to backprop pack_hook.
        self._maybe_exit_saved_tensors_hooks_scope()

    def exit_saved_tensors_hooks_scope(self):
        """
        exit saved_tensors_hooks_scope if not yet done.
        """
        assert self.entered_thread_ref() is threading.current_thread()
        if self.exit_args and not self.exited_saved_tensors_hooks_scope:
            # Note that via _register_custom_saved_tensors_hooks,
            # this saved_tensors_hooks_scope.__exit__ might get to our _custom_saved_tensors_hooks_exit below,
            # which will make sure that the order of __exit__ is correct.
            self.exited_saved_tensors_hooks_scope = True
            self.saved_tensors_hooks_scope.__exit__(*self.exit_args)

    def _pack_hook(self, x: torch.Tensor) -> Union[torch.Tensor, _GraphTensor]:
        if self.exit_args and not self.record_graph_scope.graph.is_any_recorded_tensor_alive():
            # No raw tensors alive anymore in graph_tensor_from_raw_tensor,
            # so we can exit saved_tensors_hooks_scope now.
            # (We might not always catch this properly in the Tensor _DelHook,
            #  e.g. when Tensor.__del__ runs in a different thread.)
            if _can_exit_saved_tensors_hooks_inside_hooks():
                self.exit_saved_tensors_hooks_scope()
            return x
        # _RecordGraph.__torch_dispatch__ should have recorded all newly created tensors.
        x_ = self.record_graph_scope.graph.graph_tensor_from_weak_raw_tensor.get(x, x)
        if isinstance(x_, _GraphTensor):
            x._RETURNN_grad_ckpt_del_hook = _DelHook(_WeakMethod(self._tensor_del_hook))
        return x_

    @staticmethod
    def _unpack_hook(x: Union[torch.Tensor, _GraphTensor]) -> torch.Tensor:
        if isinstance(x, _GraphTensor):
            if _can_exit_saved_tensors_hooks_inside_hooks():
                x.op.graph.gradient_checkpoint_scope_backref._maybe_exit_saved_tensors_hooks_scope()
            x.op.graph.maybe_recompute()
            return x.get_recomputed()
        return x

    def _tensor_del_hook(self):
        if _python_exit:
            return
        # Some of the relevant tensors got deleted.
        # If we are in the right thread, maybe we can do the cleanup now.
        self._maybe_exit_saved_tensors_hooks_scope()

    def _custom_saved_tensors_hooks_callback(self) -> bool:
        assert self.entered_thread_ref() is threading.current_thread()
        assert self.exit_args
        if self.record_graph_scope.graph.is_any_recorded_tensor_alive():
            return True  # keep callback alive
        else:
            self.exit_saved_tensors_hooks_scope()
            return False  # we are done, can delete callback


class _RecordGraph(TorchDispatchMode):
    def __init__(self):
        super().__init__()
        self.graph = _Graph([])

    def __torch_dispatch__(self, func, types, args=(), kwargs=None):
        kwargs = {} if kwargs is None else kwargs
        graph = self.graph
        graph.maybe_store_rng_state(torch.device("cpu"))
        graph.maybe_store_amp_state(torch.device("cpu"))
        pytree.tree_map(graph.maybe_store_rng_state, args)
        pytree.tree_map(graph.maybe_store_rng_state, kwargs)
        out = func(*args, **kwargs)
        graph.record_op(func, args, kwargs, out)
        return out


@dataclass
class _Graph:
    ops_to_be_recomputed: List[_GraphOp] = field(default_factory=list)
    graph_tensor_from_weak_raw_tensor: WeakTensorKeyDictionary[torch.Tensor, _GraphTensor] = field(
        default_factory=WeakTensorKeyDictionary
    )
    stored_device_rng_states: Dict[torch.device, Any] = field(default_factory=dict)
    stored_device_amp_states: Dict[torch.device, Any] = field(default_factory=dict)
    # Keep scope alive as long as any _GraphTensor is alive due to backprop pack_hook.
    gradient_checkpoint_scope_backref: Optional[gradient_checkpoint_scope] = None

    def is_any_recorded_tensor_alive(self) -> bool:
        """
        :return: any recorded tensor is still alive.
            Recorded tensors are outputs from any ops which were recorded,
            i.e. ops under the gradient_checkpoint_scope.
        """
        # graph_tensor_from_weak_raw_tensor is a WeakTensorKeyDictionary,
        # i.e. once there is no other strong reference to some Tensor anymore,
        # it would also be removed from graph_tensor_from_weak_raw_tensor.
        return bool(self.graph_tensor_from_weak_raw_tensor)

    def record_op(self, func: Any, args: Sequence[Any], kwargs: Dict[str, Any], out: Any):
        """record op"""
        out_flat, _ = pytree.tree_flatten(out)
        wrapped_args = pytree.tree_map_only(torch.Tensor, self.maybe_map_raw_tensor_to_graph_tensor, args)
        wrapped_kwargs = pytree.tree_map_only(torch.Tensor, self.maybe_map_raw_tensor_to_graph_tensor, kwargs)
        op = _GraphOp(
            graph=self,
            func=func,
            args=wrapped_args,
            kwargs=wrapped_kwargs,
            out_flat_num=len(out_flat),
        )
        self.ops_to_be_recomputed.append(op)
        for i, out_flat_elem in enumerate(out_flat):
            if isinstance(out_flat_elem, torch.Tensor):
                if out_flat_elem in self.graph_tensor_from_weak_raw_tensor:
                    continue
                tensor_ = _GraphTensor(op=op, out_flat_idx=i)
                self.graph_tensor_from_weak_raw_tensor[out_flat_elem] = tensor_

    def maybe_store_rng_state(self, arg: Any):
        """
        Store RNG state if not yet stored for this device.
        We store it only once for the first usage,
        as we only restore it once for the recomputation,
        and then we rely on performing the recomputation in the correct order,
        which should be deterministic and lead to the same RNG output.
        """
        if isinstance(arg, torch.Tensor):
            device = arg.device
        elif isinstance(arg, torch.device):
            device = arg
        else:
            return
        if device not in self.stored_device_rng_states:
            self.stored_device_rng_states[device] = _get_dev_rng_state(device)

    def maybe_store_amp_state(self, arg: Any):
        """store AMP state if not yet stored for this device."""
        if isinstance(arg, torch.Tensor):
            device = arg.device
        elif isinstance(arg, torch.device):
            device = arg
        else:
            return
        if device not in self.stored_device_amp_states:
            self.stored_device_amp_states[device] = _get_dev_amp_state(device)

    def maybe_map_raw_tensor_to_graph_tensor(self, tensor: torch.Tensor) -> Union[_GraphTensor, torch.Tensor]:
        """raw tensor to graph tensor if available, otherwise return raw tensor."""
        return self.graph_tensor_from_weak_raw_tensor.get(tensor, tensor)

    def maybe_recompute(self):
        """
        Recompute.

        Make sure that the recomputations happen in the correct order,
        to get any random number generator state correct.

        Note that we considered to have an API here which allowed to only recompute a subset of the ops.
        It would still compute all from op idx 0 to some given op idx, but not the rest.
        On subsequent calls, it would then continue from the last idx until again the requested op idx.
        This works fine except of one important aspect: The RNG state.
        If there are any other ops in between which use the RNG state, the RNG state would not be correct anymore.
        To allow this, we then would need to get the RNG state again and reset it later again,
        which would add some further overhead.
        To keep things simple and to avoid this overhead, we recompute all ops together right now.

        However, we can at least remove the op from the list once it is computed.
        So once any referenced tensor is not needed anymore, it can be garbage collected.
        """
        if not self.ops_to_be_recomputed:
            return
        with _reset_rng_states_scope(self.stored_device_rng_states), _reset_amp_states_scope(
            self.stored_device_amp_states
        ):
            ops_reversed_queue = list(self.ops_to_be_recomputed)
            ops_reversed_queue.reverse()
            self.ops_to_be_recomputed.clear()
            while ops_reversed_queue:
                op = ops_reversed_queue.pop(-1)
                op.recompute()
        self.stored_device_rng_states.clear()
        self.stored_device_amp_states.clear()


@dataclass
class _GraphOp:
    graph: _Graph
    func: Any
    args: Optional[Sequence[Union[_GraphTensor, Any]]]
    kwargs: Optional[Dict[str, Union[_GraphTensor, Any]]]
    out_flat_num: int
    recomputed_out_flat: Optional[Sequence[torch.Tensor]] = None

    def recompute(self):
        """recompute, assuming all args are recomputed."""
        args = pytree.tree_map_only(_GraphTensor, _GraphTensor.get_recomputed, self.args)
        kwargs = pytree.tree_map_only(_GraphTensor, _GraphTensor.get_recomputed, self.kwargs)
        out = self.func(*args, **kwargs)
        out_flat, _ = pytree.tree_flatten(out)
        assert len(out_flat) == self.out_flat_num
        self.recomputed_out_flat = out_flat
        # potentially free any referenced resources. we don't need them anymore.
        self.args = None
        self.kwargs = None
        # self.func should be ok to keep, should ref some of the low-level aten functions


@dataclass
class _GraphTensor:
    op: _GraphOp
    out_flat_idx: int

    def get_recomputed(self) -> torch.Tensor:
        """assuming it was recomputed, return the raw tensor."""
        assert self.op.recomputed_out_flat is not None
        return self.op.recomputed_out_flat[self.out_flat_idx]


@contextlib.contextmanager
def _reset_rng_states_scope(states: Dict[torch.device, Any]):
    """
    Reset RNG states scope.
    Like torch.random.fork_rng but simpler.
    """
    prev_states = {dev: _get_dev_rng_state(dev) for dev in states.keys()}
    try:
        for dev, state in states.items():
            _set_dev_rng_state(dev, state)
        yield
    finally:
        for dev, state in prev_states.items():
            _set_dev_rng_state(dev, state)


def _get_dev_rng_state(dev: torch.device):
    if dev.type == "cpu":
        return torch.get_rng_state()
    dev_mod = getattr(torch, dev.type)
    return dev_mod.get_rng_state(dev)


def _set_dev_rng_state(dev: torch.device, state: Any):
    if dev.type == "cpu":
        torch.set_rng_state(state)
    else:
        dev_mod = getattr(torch, dev.type)
        dev_mod.set_rng_state(state, dev)


@contextlib.contextmanager
def _reset_amp_states_scope(states: Dict[torch.device, Any]):
    with contextlib.ExitStack() as stack:
        for dev, state in states.items():
            if not state:
                continue
            if dev.type == "cpu":
                stack.enter_context(torch.cpu.amp.autocast(**state))
            else:
                device_module = getattr(torch, dev.type)
                stack.enter_context(device_module.amp.autocast(**state))
        yield


def _get_dev_amp_state(dev: torch.device):
    if dev.type == "cpu":
        if not torch.is_autocast_cpu_enabled():
            return None
        return {
            "dtype": torch.get_autocast_cpu_dtype(),
            "cache_enabled": torch.is_autocast_cache_enabled(),
        }

    if dev.type == "cuda":
        if not torch.is_autocast_enabled():
            return None
        return {
            "dtype": torch.get_autocast_gpu_dtype(),
            "cache_enabled": torch.is_autocast_cache_enabled(),
        }

    device_module = getattr(torch, dev.type)
    if hasattr(device_module, "is_autocast_enabled") and hasattr(device_module, "get_autocast_dtype"):
        if not device_module.is_autocast_enabled():
            return None
        return {
            "dtype": device_module.get_autocast_dtype(),
            "cache_enabled": torch.is_autocast_cache_enabled(),
        }

    return None


class _DelHook:
    def __init__(self, callback):
        self.callback = callback

    def __del__(self):
        self.callback()


class _WeakMethod:
    # wrong type hint because mypy/PyCharm don't handle MethodType well
    def __init__(self, method: Union[MethodType, Callable], *, return_if_dead: Any = None):
        assert isinstance(method, MethodType)
        self.obj = ref(method.__self__)
        self.func = method.__func__
        self.return_if_dead = return_if_dead

    def __call__(self, *args, **kwargs):
        obj = self.obj()
        if obj is None:
            return self.return_if_dead
        return self.func(obj, *args, **kwargs)


def _can_exit_saved_tensors_hooks_inside_hooks() -> bool:
    """
    Check whether we can call torch.autograd.graph.saved_tensors_hooks.__exit__
    inside the pack or unpack hook itself.

    https://github.com/rwth-i6/returnn/issues/1581
    https://github.com/pytorch/pytorch/issues/130734
    """
    # TODO Return True in some later PyTorch version? Check https://github.com/pytorch/pytorch/issues/130734.
    return False


_orig_saved_tensors_hooks_enter = torch.autograd.graph.saved_tensors_hooks.__enter__
_orig_saved_tensors_hooks_exit = torch.autograd.graph.saved_tensors_hooks.__exit__
_custom_saved_tensors_hooks_tls_ctx = threading.local()
_custom_saved_tensors_hooks_lock = threading.Lock()  # only needed for non thread-locals, i.e. threads, methods
_custom_saved_tensors_hooks_registered_threads = WeakSet()


def _register_custom_saved_tensors_hooks(*, existing_scope: torch.autograd.graph.saved_tensors_hooks):
    """
    The purpose of our custom saved_tensors_hooks __enter__/__exit__ is to make sure that
    the order of __exit__ is correct, i.e. that we exit the scope in the correct order.

    See :func:`_custom_saved_tensors_hooks_enter` and :func:`_custom_saved_tensors_hooks_exit`.

    There is no need to call :func:`_unregister_custom_saved_tensors_hooks` later.
    It will be called automatically when the last scope is exited.
    """
    thread = threading.current_thread()
    with _custom_saved_tensors_hooks_lock:
        if thread in _custom_saved_tensors_hooks_registered_threads:
            return
        if getattr(_custom_saved_tensors_hooks_tls_ctx, "stack", None) is None:
            _custom_saved_tensors_hooks_tls_ctx.stack = []
            _custom_saved_tensors_hooks_tls_ctx.in_callback = False
            _custom_saved_tensors_hooks_tls_ctx.callbacks = []
            _custom_saved_tensors_hooks_tls_ctx.queued_exits = []
        _custom_saved_tensors_hooks_tls_ctx.active = True
        _custom_saved_tensors_hooks_tls_ctx.stack.append(existing_scope)
        _custom_saved_tensors_hooks_registered_threads.add(thread)
        if len(_custom_saved_tensors_hooks_registered_threads) == 1:
            torch.autograd.graph.saved_tensors_hooks.__enter__ = _custom_saved_tensors_hooks_enter
            torch.autograd.graph.saved_tensors_hooks.__exit__ = _custom_saved_tensors_hooks_exit


def _unregister_custom_saved_tensors_hooks():
    thread = threading.current_thread()
    with _custom_saved_tensors_hooks_lock:
        assert thread in _custom_saved_tensors_hooks_registered_threads
        assert (
            not _custom_saved_tensors_hooks_tls_ctx.stack
            and (_custom_saved_tensors_hooks_tls_ctx.in_callback or not _custom_saved_tensors_hooks_tls_ctx.callbacks)
            and not _custom_saved_tensors_hooks_tls_ctx.queued_exits
        ), (
            f"_unregister_custom_saved_tensors_hooks:"
            f" stack {_custom_saved_tensors_hooks_tls_ctx.stack},"
            f" in_callback {_custom_saved_tensors_hooks_tls_ctx.in_callback},"
            f" callbacks {_custom_saved_tensors_hooks_tls_ctx.callbacks},"
            f" queued_exits {_custom_saved_tensors_hooks_tls_ctx.queued_exits}"
        )
        _custom_saved_tensors_hooks_tls_ctx.active = False
        _custom_saved_tensors_hooks_registered_threads.remove(thread)
        if not _custom_saved_tensors_hooks_registered_threads:
            torch.autograd.graph.saved_tensors_hooks.__enter__ = _orig_saved_tensors_hooks_enter
            torch.autograd.graph.saved_tensors_hooks.__exit__ = _orig_saved_tensors_hooks_exit


def _custom_saved_tensors_hooks_enter(self: torch.autograd.graph.saved_tensors_hooks):
    _custom_saved_tensors_hooks_call_callbacks()
    # The callbacks might have unregistered us. Only add to the stack if we are still active.
    if _custom_saved_tensors_hooks_tls_ctx.active:
        _custom_saved_tensors_hooks_tls_ctx.stack.append(self)
    return _orig_saved_tensors_hooks_enter(self)


def _custom_saved_tensors_hooks_exit(
    self: torch.autograd.graph.saved_tensors_hooks, exc_type=None, exc_val=None, exc_tb=None
):
    if self not in _custom_saved_tensors_hooks_tls_ctx.stack:
        raise Exception(
            f"saved_tensors_hooks __exit__ mismatch."
            f" stack {_custom_saved_tensors_hooks_tls_ctx.stack},"
            f" queued_exits {_custom_saved_tensors_hooks_tls_ctx.queued_exits},"
            f" got self {self}"
        )
    _custom_saved_tensors_hooks_tls_ctx.queued_exits.append(self)
    _custom_saved_tensors_hooks_call_callbacks()
    while _custom_saved_tensors_hooks_tls_ctx.stack:
        scope = _custom_saved_tensors_hooks_tls_ctx.stack[-1]
        if scope not in _custom_saved_tensors_hooks_tls_ctx.queued_exits:
            # Need to wait for this scope to exit first.
            # Once we exit it, we would then exit also the others when they are on top.
            break
        _custom_saved_tensors_hooks_tls_ctx.stack.pop(-1)
        _custom_saved_tensors_hooks_tls_ctx.queued_exits.remove(scope)
        _orig_saved_tensors_hooks_exit(scope, exc_type, exc_val, exc_tb)
        exc_type, exc_val, exc_tb = None, None, None  # do not propagate this again (even though it's ignored anyway)
    if not _custom_saved_tensors_hooks_tls_ctx.stack:
        assert not _custom_saved_tensors_hooks_tls_ctx.queued_exits
        if _custom_saved_tensors_hooks_tls_ctx.active:  # might have been unregistered in the meantime by callbacks
            _unregister_custom_saved_tensors_hooks()


def _register_custom_saved_tensors_hooks_thread_local_callback(cb: Callable[[], bool]):
    """
    Register some thread-local callback function which is called on saved_tensors_hooks __enter__ and __exit__.
    If it returns True, it is kept alive, otherwise removed.
    """
    assert not _custom_saved_tensors_hooks_tls_ctx.in_callback
    _custom_saved_tensors_hooks_tls_ctx.callbacks.append(cb)


def _custom_saved_tensors_hooks_call_callbacks():
    if _custom_saved_tensors_hooks_tls_ctx.in_callback:
        return  # avoid recursive calls
    try:
        _custom_saved_tensors_hooks_tls_ctx.in_callback = True
        _custom_saved_tensors_hooks_tls_ctx.callbacks = [
            cb for cb in _custom_saved_tensors_hooks_tls_ctx.callbacks if cb()
        ]
        if not _custom_saved_tensors_hooks_tls_ctx.active:  # was cleaned up by some of the callbacks
            assert not _custom_saved_tensors_hooks_tls_ctx.callbacks and not _custom_saved_tensors_hooks_tls_ctx.stack
    finally:
        _custom_saved_tensors_hooks_tls_ctx.in_callback = False


def _python_exit_handler():
    global _python_exit
    _python_exit = True


_python_exit = False
atexit.register(_python_exit_handler)



"""
This provides the Theano Op `SprintErrorSigOp` to get a loss and error signal
which is calculated via Sprint.
And there are helper classes to communicate with the Sprint subprocess
to transfer the posteriors and get back the loss and error signal.
It uses the SprintControl Sprint interface for the communication.
"""

from __future__ import annotations

import numpy
import sys
import os
import atexit
import signal
import typing
from threading import RLock, Thread
import returnn.util.task_system as task_system
from returnn.util.task_system import numpy_set_unused
from returnn.util.basic import eval_shell_str, make_hashable, close_all_fds_except
import returnn.util.basic as util
from returnn.log import log


class SprintSubprocessInstance:
    """
    The Sprint instance which is used to calculate the error signal.
    Communication is over a pipe. We pass the fds via cmd-line to the child proc.
    Basic protocol with the subprocess (encoded via pickle):
      P2C: tuple (cmd, *cmd_args). cmd is any str.
      C2P: tuple (status, *res_args). status == "ok" if no error.
    Commands:
      "init", name, version -> "ok", child_name, version
      "exit" -> (exit)
      "get_loss_and_error_signal", seg_name, seg_len, posteriors -> "ok", loss, error_signal
        Numpy arrays encoded via TaskSystem.Pickler (which is optimized for Numpy).
    On the Sprint side, we handle this via the SprintControl Sprint interface.
    """

    Version = 1  # increase when some protocol changes

    # Keep argument names as is, as these are coming directly from a user config file.
    # noinspection PyPep8Naming
    def __init__(
        self,
        sprintExecPath,
        minPythonControlVersion=2,
        sprintConfigStr="",
        sprintControlConfig=None,
        usePythonSegmentOrder=True,
    ):
        """
        :param str sprintExecPath: this executable will be called for the sub proc.
        :param int minPythonControlVersion: will be checked in the subprocess. via Sprint PythonControl
        :param str sprintConfigStr: passed to Sprint as command line args.
          can have "config:" prefix - in that case, looked up in config.
          handled via eval_shell_str(), can thus have lazy content (if it is callable, will be called).
        :param dict[str]|None sprintControlConfig: passed to SprintControl.init().
        """
        assert os.path.exists(sprintExecPath)
        self.sprintExecPath = sprintExecPath
        self.minPythonControlVersion = minPythonControlVersion
        if sprintConfigStr.startswith("config:"):
            from returnn.config import get_global_config

            config = get_global_config()
            assert config
            sprintConfigStr = config.typed_dict[sprintConfigStr[len("config:") :]]
        self.sprintConfig = eval_shell_str(sprintConfigStr)
        self.sprintControlConfig = sprintControlConfig
        self.usePythonSegmentOrder = usePythonSegmentOrder
        self.child_pid = None  # type: typing.Optional[int]
        self.parent_pid = os.getpid()
        # There is no generic way to see whether Python is exiting.
        # This is our workaround. We check for it in self.run_inner().
        self.python_exit = False
        atexit.register(self.exit_handler)
        self._cur_seg_name = None
        self._cur_posteriors_shape = None
        self.is_calculating = False
        self.init()

    def _exit_child(self, should_interrupt=False):
        if self.child_pid:
            interrupt = False
            expected_exit_status = 0 if not self.python_exit else None
            if self._join_child(wait=False, expected_exit_status=expected_exit_status) is False:  # Not yet terminated.
                interrupt = should_interrupt
                if interrupt:
                    print("SprintSubprocessInstance: interrupt child proc %i" % self.child_pid, file=log.v5)
                    os.kill(self.child_pid, signal.SIGKILL)
                else:
                    # noinspection PyBroadException
                    try:
                        self._send(("exit",))
                    except Exception:
                        pass
            else:
                self.child_pid = None
            try:
                self.pipe_p2c[1].close()
            except IOError:
                pass
            try:
                self.pipe_c2p[0].close()
            except IOError:
                pass
            if self.child_pid:
                self._join_child(wait=True, expected_exit_status=0 if not interrupt else None)
                self.child_pid = None

    # noinspection PyMethodMayBeStatic
    def _env_update_child(self):
        theano_flags = {
            key: value for (key, value) in [s.split("=", 1) for s in os.environ.get("THEANO_FLAGS", "").split(",") if s]
        }

        # First set some sane default for compile dir.
        theano_flags.setdefault(
            "compiledir_format", "compiledir_%(platform)s-%(processor)s-%(python_version)s-%(python_bitwidth)s"
        )
        compiledir_format = theano_flags["compiledir_format"]
        p = compiledir_format.find("--dev-")  # Device.startProc might have added that.
        if p >= 0:
            compiledir_format = compiledir_format[:p]
        compiledir_format += "--sprint-sub"
        theano_flags["compiledir_format"] = compiledir_format
        theano_flags["device"] = "cpu"  # Force CPU.
        theano_flags["force_device"] = True
        os.environ["THEANO_FLAGS"] = ",".join(["%s=%s" % (key, value) for (key, value) in sorted(theano_flags.items())])

    def _start_child(self):
        assert self.child_pid is None
        self.pipe_c2p = self._pipe_open()
        self.pipe_p2c = self._pipe_open()
        args = self._build_sprint_args()
        print("SprintSubprocessInstance: exec", args, file=log.v5)

        pid = os.fork()
        if pid == 0:  # child
            print("SprintSubprocessInstance: starting, pid %i" % os.getpid(), file=log.v5)
            # noinspection PyBroadException
            try:
                self._env_update_child()
                sys.stdin.close()  # Force no tty stdin.
                self.pipe_c2p[0].close()
                self.pipe_p2c[1].close()
                close_all_fds_except([0, 1, 2, self.pipe_c2p[1].fileno(), self.pipe_p2c[0].fileno()])
                os.execv(args[0], args)  # Does not return if successful.
            except BaseException:
                print("SprintSubprocessInstance: Error when starting Sprint %r." % args, file=log.v1)
                sys.excepthook(*sys.exc_info())
            finally:
                # noinspection PyUnresolvedReferences,PyProtectedMember
                os._exit(1)
                return  # Not reached.

        # parent
        self.pipe_c2p[1].close()
        self.pipe_p2c[0].close()
        self.child_pid = pid

        try:
            self._send(("init", "SprintSubprocessInstance", self.Version))
            ret = self._read()
            assert ret[0] == "ok" and len(ret) >= 3 and ret[2] == self.Version
        except Exception:
            print("SprintSubprocessInstance: Sprint child process (%r) caused an exception." % args, file=log.v1)
            sys.excepthook(*sys.exc_info())
            raise Exception("SprintSubprocessInstance Sprint init failed")

    # noinspection PyMethodMayBeStatic
    def _pipe_open(self):
        readend, writeend = os.pipe()
        if hasattr(os, "set_inheritable"):
            # https://www.python.org/dev/peps/pep-0446/
            os.set_inheritable(readend, True)
            os.set_inheritable(writeend, True)
        readend = os.fdopen(readend, "rb", 0)
        writeend = os.fdopen(writeend, "wb", 0)
        return readend, writeend

    @property
    def _my_python_mod_path(self):
        from returnn import __root_dir__

        return __root_dir__

    def _build_sprint_args(self):
        config_str = "c2p_fd:%i,p2c_fd:%i" % (self.pipe_c2p[1].fileno(), self.pipe_p2c[0].fileno())
        config_str += ",minPythonControlVersion:%i" % self.minPythonControlVersion
        if task_system.SharedMemNumpyConfig["enabled"]:
            config_str += ",EnableAutoNumpySharedMemPickling:True"
        if self.sprintControlConfig:
            config_str += "," + ",".join(["%s:%s" % (k, v) for (k, v) in sorted(self.sprintControlConfig.items())])
        my_mod_name = "returnn.sprint.control"
        args = [
            self.sprintExecPath,
            # Enable Sprint PythonControl
            "--*.python-control-enabled=true",
            # Sprint PythonControl or PythonTrainer
            "--*.pymod-path=%s" % self._my_python_mod_path,
            "--*.pymod-name=%s" % my_mod_name,
            "--*.pymod-config=%s" % config_str,
        ]
        if self.usePythonSegmentOrder:
            args += [
                # Sprint PythonSegmentOrder
                "--*.python-segment-order=true",
                "--*.python-segment-order-pymod-path=%s" % self._my_python_mod_path,
                "--*.python-segment-order-pymod-name=%s" % my_mod_name,
                "--*.python-segment-order-config=%s" % config_str,
                "--*.python-segment-order-allow-copy=false",
            ]
        args += self.sprintConfig
        return args

    def _send(self, v):
        assert os.getpid() == self.parent_pid
        p = self.pipe_p2c[1]  # see _start_child
        util.write_pickled_object(p, v)

    def _read(self):
        assert os.getpid() == self.parent_pid
        p = self.pipe_c2p[0]  # see _start_child
        return util.read_pickled_object(p)

    def _poll(self):
        assert os.getpid() == self.parent_pid
        p = self.pipe_c2p[0]  # see _start_child
        from select import select

        ready, _, _ = select([p.fileno()], [], [], 0)
        return bool(ready)

    def _join_child(self, wait=True, expected_exit_status=None):
        assert self.child_pid
        options = 0 if wait else os.WNOHANG
        pid, exit_status = os.waitpid(self.child_pid, options)
        if not wait and pid == 0:
            return False
        assert pid == self.child_pid
        if expected_exit_status is not None:
            assert exit_status == expected_exit_status, "Sprint exit code is %i" % exit_status
        return True

    def get_loss_and_error_signal__send(self, seg_name, seg_len, log_posteriors):
        """
        :param str seg_name: the segment name (seq_tag)
        :param int seg_len: the segment length in frames
        :param numpy.ndarray log_posteriors: 2d (time,label) float array, log probs
        """
        assert not self.is_calculating
        assert seg_name
        self._cur_seg_name = seg_name
        assert seg_len == log_posteriors.shape[0]
        self._cur_posteriors_shape = log_posteriors.shape
        try:
            self._send(("get_loss_and_error_signal", seg_name, seg_len, log_posteriors.astype("float32", copy=False)))
        except (IOError, EOFError):
            raise
        else:
            self.is_calculating = True

    def get_loss_and_error_signal__have_data(self):
        """
        :rtype: bool
        """
        assert self.is_calculating
        return self._poll()

    def get_loss_and_error_signal__read(self):
        """
        :rtype (str, float, numpy.ndarray)
        :returns (seg_name, loss, error_signal). error_signal has the same shape as posteriors.
        """
        assert self.is_calculating
        try:
            ret = self._read()
        except (IOError, EOFError):
            raise
        else:
            self.is_calculating = False
        assert ret[0] == "ok" and len(ret) == 3, "Got unexpected return: %r" % (ret,)
        loss = ret[1]
        error_signal = ret[2]
        assert error_signal.shape == self._cur_posteriors_shape
        return self._cur_seg_name, loss, error_signal

    def exit_handler(self):
        """
        Called at exit. Exit child.
        """
        assert os.getpid() == self.parent_pid
        self.python_exit = True
        self._exit_child(should_interrupt=True)

    def init(self):
        """
        Init. (Re)Start child.
        """
        self._exit_child()
        self._start_child()


class ReaderThread(Thread):
    """
    Sprint reader thread.
    """

    def __init__(
        self, instance, instance_idx, batch_idxs, tags, seq_lengths, log_posteriors, batch_loss, batch_error_signal
    ):
        """
        :param SprintSubprocessInstance instance:
        :param int instance_idx:
        :param list[int] batch_idxs:
        :param list[str] tags: seq names, length = batch
        :param numpy.ndarray seq_lengths: 1d (batch)
        :param numpy.ndarray log_posteriors: 3d (time,batch,label)
        :param numpy.ndarray batch_loss: 1d (batch). will write result into it.
        :param numpy.ndarray batch_error_signal: 3d (time,batch,label). will write results into it.
        """
        super(ReaderThread, self).__init__(
            name="SprintErrorSignals reader thread for Sprint instance %i" % instance_idx
        )
        self.daemon = True
        self.instance_idx = instance_idx
        self.instance = instance
        self.batch_idxs = batch_idxs
        self.tags = tags
        self.seq_lengths = seq_lengths
        self.log_posteriors = log_posteriors
        self.batch_loss = batch_loss
        self.batch_error_signal = batch_error_signal
        self.exception = None
        self.start()

    def run(self):
        """
        Main thread func.
        """
        try:
            for b in self.batch_idxs:
                self.instance.get_loss_and_error_signal__send(
                    seg_name=self.tags[b],
                    seg_len=self.seq_lengths[b],
                    log_posteriors=self.log_posteriors[: self.seq_lengths[b], b],
                )
                seg_name, loss, error_signal = self.instance.get_loss_and_error_signal__read()
                assert seg_name == self.tags[b]
                self.batch_loss[b] = loss
                self.batch_error_signal[: self.seq_lengths[b], b] = error_signal
                numpy_set_unused(error_signal)
        except Exception as exc:
            self.exception = exc


class SprintInstancePool:
    """
    This is a pool of Sprint instances.
    First, for each unique sprint_opts, there is a singleton
      which can be accessed via get_global_instance.
    Then, this can be used in multiple ways.
      (1) get_batch_loss_and_error_signal.
      (2) ...
    """

    class_lock = RLock()
    global_instances = {}  # sprint_opts -> SprintInstancePool instance

    @classmethod
    def get_global_instance(cls, sprint_opts):
        """
        :param dict[str] sprint_opts:
        :rtype: SprintInstancePool
        """
        sprint_opts = make_hashable(sprint_opts)
        with cls.class_lock:
            if sprint_opts in cls.global_instances:
                return cls.global_instances[sprint_opts]
            instance = SprintInstancePool(sprint_opts=sprint_opts)
            cls.global_instances[sprint_opts] = instance
            return instance

    def __init__(self, sprint_opts):
        """
        :param dict[str] sprint_opts:
        """
        # The lock will not be acquired automatically on the public functions here as there is the valid
        # usage that only one thread will access it anyway.
        # So, take care of acquiring this lock yourself whenever you call here potentially from multiple threads.
        # All the code is not thread-safe, so this is important!
        self.lock = RLock()
        assert isinstance(sprint_opts, dict)
        sprint_opts = sprint_opts.copy()
        self.max_num_instances = int(sprint_opts.pop("numInstances", 1))
        self.sprint_opts = sprint_opts
        self.instances = []  # type: typing.List[SprintSubprocessInstance]

    def _maybe_create_new_instance(self):
        """
        :rtype: SprintSubprocessInstance|None
        """
        if len(self.instances) < self.max_num_instances:
            self.instances.append(SprintSubprocessInstance(**self.sprint_opts))
            return self.instances[-1]
        return None

    def _get_instance(self, i):
        """
        :param int i:
        :rtype: SprintSubprocessInstance
        """
        assert i < self.max_num_instances
        if i >= len(self.instances):
            assert i == len(self.instances)
            self._maybe_create_new_instance()
        return self.instances[i]

    def get_batch_loss_and_error_signal(self, log_posteriors, seq_lengths, tags=None):
        """
        :param numpy.ndarray log_posteriors: 3d (time,batch,label)
        :param numpy.ndarray seq_lengths: 1d (batch)
        :param list[str] tags: seq names, length = batch
        :rtype (numpy.ndarray, numpy.ndarray)
        :returns (loss, error_signal). error_signal has the same shape as posteriors.
        loss is a 1d-array (batch).

        Note that this accesses some global references, like global current seg info,
        via the current Device instance.
        Thus this is expected to be run from the Device host proc,
          inside from SprintErrorSigOp.perform.
        This also expects that we don't have chunked seqs.
        """
        assert seq_lengths.ndim == 1
        assert log_posteriors.ndim == 3
        n_batch = seq_lengths.shape[0]
        assert n_batch == log_posteriors.shape[1]

        if tags is None:
            raise NotImplementedError("This feature was removed with dropped Theano support")
        assert len(tags) == n_batch

        batch_loss = numpy.zeros((n_batch,), dtype="float32")
        batch_error_signal = numpy.zeros_like(log_posteriors, dtype="float32")

        # greedy solution to the scheduling problem
        sorted_length = sorted(enumerate(seq_lengths), key=lambda x: x[1], reverse=True)
        jobs = [[] for _ in range(self.max_num_instances)]
        joblen = [0] * self.max_num_instances
        for i, l in sorted_length:
            j = min(enumerate(joblen), key=lambda x: x[1])[0]  # noqa
            jobs[j].append(i)
            joblen[j] += l

        if self.max_num_instances > 1:
            threads = [
                ReaderThread(
                    self._get_instance(i), i, jobs[i], tags, seq_lengths, log_posteriors, batch_loss, batch_error_signal
                )
                for i in range(self.max_num_instances)
            ]
            for i, thread in enumerate(threads):
                thread.join()
                if thread.exception:
                    raise thread.exception
        else:
            # Very simple parallelism. We must avoid any form of multi-threading
            # because this can be problematic with Theano.
            # See: https://groups.google.com/forum/#!msg/theano-users/Pu4YKlZKwm4/eNcAegzaNeYJ
            # We also try to keep it simple here.
            for bb in range(0, n_batch, self.max_num_instances):
                for i in range(self.max_num_instances):
                    b = bb + i
                    if b >= n_batch:
                        break
                    instance = self._get_instance(i)
                    instance.get_loss_and_error_signal__send(
                        seg_name=tags[b], seg_len=seq_lengths[b], log_posteriors=log_posteriors[: seq_lengths[b], b]
                    )
                for i in range(self.max_num_instances):
                    b = bb + i
                    if b >= n_batch:
                        break
                    instance = self._get_instance(i)
                    seg_name, loss, error_signal = instance.get_loss_and_error_signal__read()
                    assert seg_name == tags[b]
                    batch_loss[b] = loss
                    batch_error_signal[: seq_lengths[b], b] = error_signal
                    numpy_set_unused(error_signal)
        return batch_loss, batch_error_signal

    def get_automata_for_batch(self, tags):
        """
        :param list[str]|numpy.ndarray tags: sequence names, used for Sprint (ndarray of shape (batch, max_str_len))
        :return: (edges, weights, start_end_states). all together in one automaton.
          edges are of shape (4, num_edges), each (from, to, emission-idx, seq-idx), of dtype uint32.
          weights are of shape (num_edges,), of dtype float32.
          start_end_states are of shape (2, batch), each (start,stop) state idx, batch = len(tags), of dtype uint32.
        :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)
        """
        all_num_states = [None] * len(tags)  # type: typing.List[typing.Optional[int]]
        all_num_edges = [None] * len(tags)  # type: typing.List[typing.Optional[int]]
        all_edges = [None] * len(tags)  # type: typing.List[typing.Optional[numpy.ndarray]]
        all_weights = [None] * len(tags)  # type: typing.List[typing.Optional[numpy.ndarray]]
        for bb in range(0, len(tags), self.max_num_instances):
            for i in range(self.max_num_instances):
                b = bb + i
                if b >= len(tags):
                    break
                instance = self._get_instance(i)
                if isinstance(tags[0], str):
                    segment_name = tags[b]
                elif isinstance(tags[0], bytes):
                    segment_name = tags[b].decode()
                else:
                    segment_name = tags[b].view("S%d" % tags.shape[1])[0]
                assert isinstance(segment_name, str)
                # noinspection PyProtectedMember
                instance._send(("export_allophone_state_fsa_by_segment_name", segment_name))
            for i in range(self.max_num_instances):
                b = bb + i
                if b >= len(tags):
                    break
                instance = self._get_instance(i)
                # noinspection PyProtectedMember
                r = instance._read()
                if r[0] != "ok":
                    raise RuntimeError(r[1])
                num_states, num_edges, edges, weights = r[1:]
                all_num_states[b] = num_states
                all_num_edges[b] = num_edges
                all_edges[b] = edges.reshape((3, num_edges))  # (from, to, emission-idx) for each edge, uint32
                all_weights[b] = weights  # for each edge, float32
        state_offset = 0
        for idx in range(len(all_edges)):
            num_edges = all_num_edges[idx]
            all_edges[idx][0:2, :] += state_offset
            state_offset += all_num_states[idx]
            # add sequence_idx. becomes (from, to, emission-idx, seq-idx) for each edge
            all_edges[idx] = numpy.vstack((all_edges[idx], numpy.ones((1, num_edges), dtype="uint32") * idx))

        start_end_states = numpy.empty((2, len(all_num_states)), dtype="uint32")
        state_offset = 0
        for idx, num_states in enumerate(all_num_states):
            start_end_states[0, idx] = state_offset
            start_end_states[1, idx] = state_offset + num_states - 1
            state_offset += num_states

        return numpy.hstack(all_edges), numpy.hstack(all_weights), start_end_states

    def get_free_instance(self):
        """
        :rtype: SprintSubprocessInstance|None
        """
        for inst in self.instances:
            if not inst.is_calculating:
                return inst
        return self._maybe_create_new_instance()



"""
torch.distributed utils
"""

from __future__ import annotations
import ast
import logging
import os
import socket
from typing import Optional, Any, Dict

import torch
from torch.nn.parallel import DistributedDataParallel

from returnn.config import Config
from returnn.util.basic import CollectionReadCheckCovered

_logger = logging.getLogger("returnn.torch.distributed")


class DistributedContext:
    """
    This class setups some helper functions for torch distributed training
    """

    def __init__(self, options: Dict[str, Any]):
        self._opts = CollectionReadCheckCovered(options)

        # Subprocesses have issues initializing torch.distributed process groups.
        #
        # We therefore pass rank/size information of the process group via an env
        # variable that is automatically inherited in any created subprocess.
        env_var_name = "_RETURNN_TORCH_DISTRIBUTED_INIT_INFO"
        prev_init_info = os.environ.get(env_var_name)
        if prev_init_info:
            self.prev_init_info = ast.literal_eval(prev_init_info)
            self._rank = self.prev_init_info["rank"]
            self._size = self.prev_init_info["size"]
        else:
            import torch.distributed as dist

            # When no backend is specified, we set gloo for CPU tensors and nccl for CUDA tensors as backend.
            # torch 2.6.0 and onwards require explicitly setting the backends.
            # See https://github.com/rwth-i6/returnn/issues/1724 for discussion.
            dist.init_process_group(backend=self._opts.get("backend", "cpu:gloo,cuda:nccl"))
            self._rank = dist.get_rank()
            self._size = dist.get_world_size()
            os.environ[env_var_name] = repr({"rank": self._rank, "size": self._size})

        self._local_rank = int(os.environ["LOCAL_RANK"])
        self._local_size = int(os.environ["LOCAL_WORLD_SIZE"])

        _logger.info(
            "Torch distributed initialized. Hostname %s, pid %i, rank %i / size %i, local rank %s / local size %s."
            % (socket.gethostname(), os.getpid(), self._rank, self._size, self._local_rank, self._local_size)
        )

        self._reduce_type = self._opts.get("reduce_type", "grad")
        self._param_sync_step: Optional[int] = self._opts.get("param_sync_step", None)
        if self._reduce_type == "param":
            assert isinstance(self._param_sync_step, int) and self._param_sync_step > 0, (
                f"reduce_type param: param_sync_step must be a positive int,"
                f" got {self._param_sync_step!r} ({type(self._param_sync_step).__name__})"
            )
            _logger.info(f"reduce_type param: param_sync_step {self._param_sync_step}")
        elif self._reduce_type == "grad":
            _logger.info("reduce_type grad")
        else:
            raise ValueError(f"invalid reduce_type {self._reduce_type!r}")

        self._check_no_unknown_opts()

    def __repr__(self):
        return f"<{self.__class__.__name__} size={self._size} reduce_type={self._reduce_type}>"

    def _check_no_unknown_opts(self):
        # We check that all opts in self._opts have been used.
        # This function here is called at the end in __init__,
        # and not all opts are used yet, so read them now,
        # such that the check in the end works.
        self._opts.get("backend")
        if self._reduce_type == "grad":
            self._opts.get("class")
            self._opts.get("options")
        if self._reduce_type == "param":
            self._opts.get("sync_on_cpu")

        self._opts.assert_all_read()

    def local_rank(self) -> int:
        """local rank"""
        return self._local_rank

    def local_size(self) -> int:
        """local size"""
        return self._local_size

    def rank(self) -> int:
        """global rank"""
        return self._rank

    def size(self) -> int:
        """global size"""
        return self._size

    def get_param_sync_step(self) -> Optional[int]:
        """param sync step"""
        return self._param_sync_step

    def maybe_make_distributed_module(self, module: torch.nn.Module) -> Optional[DistributedDataParallel]:
        """
        Maybe make a wrapped distributed module.

        :param module: original module
        :return: potentially wrapped module
        """
        if self._reduce_type == "param":
            return None
        assert self._reduce_type == "grad"
        cls = self._opts.get("class", DistributedDataParallel)
        if cls is not DistributedDataParallel:
            _logger.warning(f"Using custom class {cls} instead of DistributedDataParallel, might be unsupported.")
        kwargs = self._opts.get("options", {})
        return cls(
            module=module,
            device_ids=[self.local_rank()],
            **kwargs,
        )

    def should_sync_now(self, *, epoch_step_idx: int) -> bool:
        """
        :param epoch_step_idx: current step index
        :return: whether to sync the training processes in this step
        """
        if self._reduce_type == "grad":
            return True
        elif self._reduce_type == "param":
            return (epoch_step_idx % self._param_sync_step) == (self._param_sync_step - 1)
        else:
            raise ValueError(f"invalid reduce_type {self._reduce_type}")

    def step_after_param_update(self, *, module: torch.nn.Module, epoch_step_idx: int):
        """one train step"""
        if self._reduce_type == "param" and self.should_sync_now(epoch_step_idx=epoch_step_idx):
            _sync_params_avg(module=module, sync_on_cpu=self._opts.get("sync_on_cpu", False))


_is_set_up = False
_ctx = None  # type: Optional[DistributedContext]


def get_ctx(config: Optional[Config] = None) -> Optional[DistributedContext]:
    """
    :param config:
    :returns: the global context if Torch distributed is enabled, or None otherwise.
      If we did not setup the context yet, it will automatically create it.
    """
    global _is_set_up, _ctx
    if _is_set_up:
        return _ctx
    if not config:
        from returnn.config import get_global_config

        config = get_global_config(raise_exception=False)
        if not config:
            return None

    _is_set_up = True
    opts = config.typed_value("torch_distributed")
    if opts is None:
        return None

    assert isinstance(opts, dict)
    _ctx = DistributedContext(opts)
    return _ctx


@torch.no_grad()
def _sync_params_avg(*, module: torch.nn.Module, sync_on_cpu: bool = False):
    import torch.distributed as dist

    if sync_on_cpu:
        for param in module.parameters():
            # Separately move each param to CPU (instead of the whole module), to safe CPU memory.
            param_cpu = param.to(torch.device("cpu"))
            # On CPU, we are likely using Gloo, and Gloo does not support AVG
            dist.all_reduce(param_cpu.data, op=dist.ReduceOp.SUM)
            param_cpu.data /= dist.get_world_size()
            param.data = param_cpu.to(param.device)
        return

    if dist.get_backend() == "gloo":
        # Gloo does not support AVG
        reduce_op = dist.ReduceOp.SUM
    else:
        if hasattr(dist.ReduceOp, "AVG"):
            reduce_op = dist.ReduceOp.AVG
        else:
            # Older PyTorch versions do not have ReduceOp.AVG.
            reduce_op = dist.ReduceOp.SUM

    for param in module.parameters():
        dist.all_reduce(param.data, op=reduce_op)
        if reduce_op == dist.ReduceOp.SUM:
            param.data /= dist.get_world_size()



"""
:func:`lru_cache`, copied from Python functools, slightly adapted,
and extended by functions to check whether some key is cached or not.
"""

from __future__ import annotations
from typing import Dict, Any
from functools import update_wrapper
from threading import RLock
from collections import namedtuple


def lru_cache(maxsize: int = 128, typed: bool = False):
    """Least-recently-used cache decorator.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    If *typed* is True, arguments of different types will be cached separately.
    For example, f(3.0) and f(3) will be treated as distinct calls with
    distinct results.

    Arguments to the cached function must be hashable.

    Use f.cache_len() to see the current size of the cache.
    Use f.cache_set(*args, result, **kwargs) to set a value in the cache directly.
    Use f.cache_peek(*args, update_statistics=False, fallback=None, **kwargs)
    to peek the cache, without ever calling the user function.
    View the cache statistics named tuple (hits, misses, maxsize, currsize)
    with f.cache_info().
    Clear the cache and statistics with f.cache_clear().
    Remove the oldest entry from the cache with f.cache_pop_oldest().
    Take out some entry from the cache with f.cache_pop(*args, fallback=not_specified, **kwargs).
    Set the maximum cache size to a new value with f.cache_set_maxsize(new_maxsize).
    Access the underlying function with f.__wrapped__.

    See:  https://en.wikipedia.org/wiki/Cache_replacement_policies#Least_recently_used_(LRU)

    """

    # Users should only access the lru_cache through its public API:
    #       cache_info, cache_clear, and f.__wrapped__
    # The internals of the lru_cache are encapsulated for thread safety and
    # to allow the implementation to change (including a possible C version).

    if isinstance(maxsize, int):
        assert maxsize >= 0
    elif callable(maxsize) and isinstance(typed, bool):
        # The user_function was passed in directly via the maxsize argument
        user_function, maxsize = maxsize, 128
        return _lru_cache_wrapper(user_function, maxsize, typed)
    elif maxsize is not None:
        raise TypeError("Expected first argument to be an integer, a callable, or None")

    # noinspection PyShadowingNames
    def _decorating_function(user_function):
        return _lru_cache_wrapper(user_function, maxsize, typed)

    return _decorating_function


_CacheInfo = namedtuple("CacheInfo", ["hits", "misses", "maxsize", "currsize"])


def _lru_cache_wrapper(user_function, maxsize: int, typed: bool):
    # Constants shared by all lru cache instances:
    make_key = _make_key  # build a key from the function arguments
    # noinspection PyPep8Naming
    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3  # names for the link fields

    cache: Dict[Any, list] = {}
    hits = misses = 0
    full = False
    cache_get = cache.get  # bound method to lookup a key or return None
    cache_len = cache.__len__  # get cache size without calling len()
    lock = RLock()  # because linkedlist updates aren't threadsafe
    root = []  # root of the circular doubly linked list
    root[:] = [root, root, None, None]  # initialize by pointing to self

    assert maxsize >= 0

    def wrapper(*args, **kwds):
        """
        User-facing wrapper function.
        """
        # Size limited caching that tracks accesses by recency
        nonlocal root, hits, misses, full
        key = make_key(args, kwds, typed)
        with lock:
            link = cache_get(key)
            if link is not None:
                # Move the link to the front of the circular queue
                link_prev, link_next, _key, result = link
                link_prev[NEXT] = link_next
                link_next[PREV] = link_prev
                last = root[PREV]
                last[NEXT] = root[PREV] = link
                link[PREV] = last
                link[NEXT] = root
                hits += 1
                return result
            misses += 1
        result = user_function(*args, **kwds)
        if maxsize > 0:
            _cache_insert(key, result)
        return result

    def _cache_insert(key, result):
        nonlocal root, full
        with lock:
            if key in cache:
                # Getting here means that this same key was added to the
                # cache while the lock was released.  Since the link
                # update is already done, we need only return the
                # computed result and update the count of misses.
                pass
            elif full:
                # Use the old root to store the new key and result.
                oldroot = root
                oldroot[KEY] = key
                oldroot[RESULT] = result
                # Empty the oldest link and make it the new root.
                # Keep a reference to the old key and old result to
                # prevent their ref counts from going to zero during the
                # update. That will prevent potentially arbitrary object
                # clean-up code (i.e. __del__) from running while we're
                # still adjusting the links.
                root = oldroot[NEXT]
                oldkey = root[KEY]
                root[KEY] = root[RESULT] = None
                # Now update the cache dictionary.
                del cache[oldkey]
                # Save the potentially reentrant cache[key] assignment
                # for last, after the root and links have been put in
                # a consistent state.
                cache[key] = oldroot
            else:
                # Put result in a new link at the front of the queue.
                last = root[PREV]
                link = [last, root, key, result]
                last[NEXT] = root[PREV] = cache[key] = link
                # Use the cache_len bound method instead of the len() function
                # which could potentially be wrapped in an lru_cache itself.
                full = cache_len() >= maxsize

    def cache_info():
        """Report cache statistics"""
        with lock:
            return _CacheInfo(hits, misses, maxsize, cache_len())

    def cache_clear():
        """Clear the cache and cache statistics"""
        nonlocal hits, misses, full
        with lock:
            for link in cache.values():
                link.clear()  # make GC happy
            cache.clear()
            root[:] = [root, root, None, None]
            hits = misses = 0
            full = False

    def cache_parameters():
        """
        :return: parameters (maxsize, typed) of the cache as dict
        """
        return {"maxsize": maxsize, "typed": typed}

    def cache_set(*args, result, **kwargs):
        """
        Sets a value in the cache directly.
        """
        nonlocal root, full
        if maxsize > 0:
            key = make_key(args, kwargs, typed)
            _cache_insert(key, result)

    def cache_peek(*args, update_statistics: bool = True, fallback=None, **kwargs):
        """
        Peeks the cache without ever calling the user function.
        """
        nonlocal hits, misses
        key = make_key(args, kwargs, typed)
        with lock:
            link = cache_get(key)
            if link is not None:
                if update_statistics:
                    hits += 1
                return link[RESULT]
            if update_statistics:
                misses += 1
            return fallback

    not_specified = object()

    def cache_pop(*args, fallback=not_specified, **kwargs):
        """
        Removes the entry from the cache.
        """
        nonlocal hits, misses, full
        key = make_key(args, kwargs, typed)
        with lock:
            link = cache_get(key)
            if link is not None:
                # Take out link.
                link[PREV][NEXT] = link[NEXT]
                link[NEXT][PREV] = link[PREV]
                oldkey = link[KEY]
                oldvalue = link[RESULT]
                link.clear()
                del cache[oldkey]
                full = cache_len() >= maxsize
                return oldvalue
            if fallback is not_specified:
                raise KeyError("key not found")
            return fallback

    def cache_pop_oldest(*, fallback=not_specified):
        """
        Removes the oldest entry from the cache.
        """
        nonlocal root, full
        with lock:
            if not cache:
                if fallback is not_specified:
                    raise KeyError("cache is empty")
                return fallback
            assert cache
            # Take out oldest link.
            link: list = root[NEXT]
            link[NEXT][PREV] = root
            root[NEXT] = link[NEXT]
            oldkey = link[KEY]
            oldvalue = link[RESULT]
            link.clear()
            del cache[oldkey]
            full = cache_len() >= maxsize
            return oldvalue

    def cache_set_maxsize(new_maxsize: int):
        """
        Resets the maxsize.
        If the new maxsize is smaller than the current cache size, the oldest entries are removed.
        """
        nonlocal maxsize, full
        assert new_maxsize >= 0
        with lock:
            maxsize = new_maxsize
            while cache_len() > maxsize:
                cache_pop_oldest()
            full = cache_len() >= maxsize

    wrapper.cache_info = cache_info
    wrapper.cache_clear = cache_clear
    wrapper.cache_parameters = cache_parameters
    wrapper.cache_set = cache_set
    wrapper.cache_peek = cache_peek
    wrapper.cache_pop = cache_pop
    wrapper.cache_len = cache_len
    wrapper.cache_pop_oldest = cache_pop_oldest
    wrapper.cache_set_maxsize = cache_set_maxsize

    update_wrapper(wrapper, user_function)

    return wrapper


def _make_key(args, kwds, typed, *, _kwd_mark=(object(),), _fasttypes=(int, str), _tuple=tuple, _type=type, _len=len):
    """Make a cache key from optionally typed positional and keyword arguments

    The key is constructed in a way that is flat as possible rather than
    as a nested structure that would take more memory.

    If there is only a single argument and its data type is known to cache
    its hash value, then that argument is returned without a wrapper.  This
    saves space and improves lookup speed.

    """
    # All of code below relies on kwds preserving the order input by the user.
    # Formerly, we sorted() the kwds before looping.  The new way is *much*
    # faster; however, it means that f(x=1, y=2) will now be treated as a
    # distinct call from f(y=2, x=1) which will be cached separately.
    key = args
    if kwds:
        key += _kwd_mark
        for item in kwds.items():
            key += item
    if typed:
        key += _tuple(_type(v) for v in args)  # noqa
        if kwds:
            key += _tuple(_type(v) for v in kwds.values())  # noqa
    elif _len(key) == 1 and _type(key[0]) in _fasttypes:
        return key[0]
    return _HashedSeq(key)


class _HashedSeq(list):
    """
    This class guarantees that hash() will be called no more than once
    per element.  This is important because the lru_cache() will hash
    the key multiple times on a cache miss.
    """

    __slots__ = ("hashvalue",)

    def __init__(self, tup, *, _hash=hash):
        super().__init__(tup)
        self.hashvalue = _hash(tup)

    def __hash__(self):
        return self.hashvalue



"""
stft etc
"""

from __future__ import annotations
from typing import Optional, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["stft"]


def stft(
    x: Tensor,
    *,
    in_spatial_dim: Dim,
    frame_step: int,
    frame_length: int,
    fft_length: Optional[int] = None,
    window_use_frame_length: bool = True,
    align_window_left: bool = True,
    window_enforce_even: bool = True,
    out_spatial_dim: Optional[Dim] = None,
    out_dim: Optional[Dim] = None,
    use_mask: Optional[bool] = None,
) -> Tuple[Tensor, Dim, Dim]:
    """
    Calculate the short-time Fourier transform (STFT) of a signal.

    We currently always use the Hann window.

    Note that there are inconsistencies between frameworks (e.g. PyTorch vs TensorFlow).
    We have options where you can explicitly specify the behavior,
    to match either PyTorch or TensorFlow.
    See here for some discussion and demonstration:
    https://github.com/pytorch/pytorch/issues/100177
    https://github.com/librosa/librosa/issues/1701
    https://github.com/albertz/playground/blob/master/tf_pt_stft.py

    :param x: (..., in_spatial_dim, ...). any other dims are treated as batch dims.
    :param in_spatial_dim:
    :param frame_length: win_length in PT/librosa, frame_size in StftLayer, frame_length in TF
    :param frame_step: hop_length in PT/librosa, frame_shift in StftLayer, frame_step in TF
    :param fft_length: n_fft in PT/librosa, fft_size in StftLayer, fft_length in TF.
        If not given, will use frame_length.
    :param window_use_frame_length: Whether to use window size = frame_length.
        If False, uses window size = fft_length.
        This has only an effect if frame_length != fft_length.
        This has an effect on the output seq length.
        If True: out_seq_len = ⌈(T - frame_length + 1) / frame_step⌉ (T = in_seq_len).
        If False: out_seq_len = ⌈(T - fft_length + 1) / frame_step⌉.
        Note that in TF/SciPy, the behavior matches to window_use_frame_length=True,
        but in PT/librosa, the behavior matches to window_use_frame_length=False.
    :param align_window_left: whether to align the window to the left inside the potential wider fft_length window.
        This only has an effect if frame_length != fft_length.
        If True, the window will be aligned to the left and the right side will be padded with zeros.
        Then effectively the remaining phases are not used.
        If False, the window will be aligned to the center and the left and right side will be padded with zeros.
        Then effectively the beginning and end phases are not used.
        In TF/SciPy, the behavior matches to align_window_left=True,
        but in PT/librosa, the behavior matches to align_window_left=False.
    :param window_enforce_even: enforce that the window size is even, potentially lowering the window size by 1.
        This only has an effect if the window size is uneven.
        In TF, the behavior matches to window_enforce_even=True,
        but in most other frameworks, the behavior matches to window_enforce_even=False.
    :param out_spatial_dim:
    :param out_dim:
    :param use_mask:
    :return: (stft, out_spatial_dim, out_dim)
    """
    if in_spatial_dim.need_masking():
        if use_mask is None:
            use_mask = rf.use_mask_default(default=True, default_false_for_behavior_version_up_to=22)
        if use_mask:
            x = x.copy_masked(0, dims=[in_spatial_dim])
    fft_length = fft_length or frame_length
    if out_dim is None:
        out_dim = Dim(fft_length // 2 + 1, name="stft-freq")
    if out_spatial_dim is None:
        from .conv import make_conv_out_spatial_dims

        (out_spatial_dim,) = make_conv_out_spatial_dims(
            [in_spatial_dim], filter_size=frame_length, strides=frame_step, padding="valid"
        )
    # noinspection PyProtectedMember
    return (
        x._raw_backend.stft(
            x,
            in_spatial_dim=in_spatial_dim,
            frame_step=frame_step,
            frame_length=frame_length,
            fft_length=fft_length,
            window_use_frame_length=window_use_frame_length,
            align_window_left=align_window_left,
            window_enforce_even=window_enforce_even,
            out_spatial_dim=out_spatial_dim,
            out_dim=out_dim,
        ),
        out_spatial_dim,
        out_dim,
    )



"""
The :class:`Tensor` class represents a tensor.
This is framework-agnostic, i.e. is fine for TensorFlow, PyTorch, or any other framework.
(Earlier, this was the class ``Data`` in RETURNN, and ``Data`` is now an alias for this class.)

This class is to describe a tensor,
i.e. its shape and properties like
whether we should consider it sparse data (i.e. it represents indices).
Each dimension is described by :class:`Dim`.

This is used in :class:`TFNetwork` to describe the dataset external data (:class:`ExternData`)
as well as in every layer's output and in many other parts of the code.

See :ref:`data`.

This is designed in a way that is efficient for eager-mode execution such as PyTorch,
but at the same time compatible with older RETURNN code.

Discussion on the move from loosely TF-specific ``Data`` to framework-agnostic ``Tensor``:
https://github.com/rwth-i6/returnn/issues/1165
"""

from __future__ import annotations
from typing import Optional, Union, Sequence, Tuple, Generic, TypeVar, Set

from returnn.util.basic import NotSpecified
from .dim import Dim
from ._tensor_extra import _TensorExtra, _TensorMixin
from ._tensor_op_overloads import _TensorOpOverloadsMixin
import returnn.tensor._tensor_extra as _tensor_extra


__all__ = ["Tensor"]


RawTensorType = TypeVar("RawTensorType")  # e.g. torch.Tensor, tf.Tensor, numpy.ndarray, ...


class Tensor(_TensorMixin, _TensorOpOverloadsMixin, Generic[RawTensorType]):
    """
    Represents a tensor, in a frame-agnostic way. See the module docstring.
    """

    size_dtype = "int32"

    __slots__ = ("name", "_dims", "dtype", "sparse_dim", "_raw_tensor", "_feature_dim_axis", "version", "_extra")

    name: str
    _dims: Tuple[Dim, ...]
    dtype: str
    sparse_dim: Optional[Dim]
    _raw_tensor: Optional[RawTensorType]

    _feature_dim_axis: Optional[Union[int, NotSpecified]]  # https://github.com/rwth-i6/returnn/issues/1273
    version: int
    _extra: Optional[_TensorExtra]

    # This is potentially replaced by native implementation.
    def __init__(
        self,
        name: str,
        dims: Optional[Sequence[Dim]] = None,
        dtype: Optional[str] = None,
        *,
        sparse_dim: Optional[Dim] = None,
        feature_dim: Optional[Dim] = None,
        feature_dim_axis: Optional[Union[int, NotSpecified]] = NotSpecified,
        raw_tensor: Optional[RawTensorType] = None,
        version: Optional[int] = None,
        **kwargs,
    ):
        """
        :param name:
        :param dims: the shape, where each dimension is described by a :class:`Dim`.
        :param dtype: e.g. "float32" or "int64"
        :param sparse_dim: when the values are indices into some dimension, this is the dimension.
            You can also interpret the whole tensor as a sparse representation of a dense one-hot tensor,
            where this sparse_dim becomes the additional dense dimension.
        :param raw_tensor: the raw tensor, e.g. numpy array, TF tensor, or PyTorch tensor
        :param version: behavior version just for Tensor. If not specified, and `dims` is None (old code),
            it uses version 1.
            - v1: the old behavior of Data. Specifically, time_dim_axis and feature_dim_axis are used
                and automatically inferred when not specified.
            - v2: time_dim_axis, feature_dim_axis are None by default.
        :param kwargs: see :func:`_handle_extra_kwargs`, :func:`infer_dim_tags`
        """
        if ("sparse" in kwargs or "vocab" in kwargs) and sparse_dim is None:
            sparse_dim = _tensor_extra.infer_sparse_dim(name=name, **kwargs)
        if dims is not None:
            assert "shape" not in kwargs and "dim_tags" not in kwargs  # probably old code got this wrong
            if version is None:
                version = 2
        else:
            # old code
            dims = _tensor_extra.infer_dim_tags(
                name=name, sparse_dim=sparse_dim, feature_dim_axis=feature_dim_axis, **kwargs
            )
            if version is None:
                version = 1
        if dtype is None:
            # old defaults
            if version == 1:
                dtype = "int32" if sparse_dim else "float32"
            else:
                raise ValueError("Tensor dtype needs to be specified")

        self.name = name
        self._dims = tuple(dims)
        self.dtype = dtype
        self.sparse_dim = sparse_dim
        self._raw_tensor = None  # assignment below
        self.version = version
        self._extra = None  # type: Optional[_TensorExtra]

        if feature_dim_axis is NotSpecified:
            # version == 1 leaves it as NotSpecified,
            #   which enables some dynamic default behavior, like taking the last axis.
            # version >= 2 avoids any default behavior, and just sets it to None.
            if version >= 2:
                feature_dim_axis = None
        elif feature_dim_axis is None:
            pass
        elif isinstance(feature_dim_axis, int):
            assert not self.sparse_dim, "cannot have feature_dim_axis when sparse"
            if feature_dim_axis < 0:
                feature_dim_axis += self.batch_ndim
            assert 0 <= feature_dim_axis < self.batch_ndim
        else:
            raise TypeError(f"unexpected feature_dim_axis type {type(feature_dim_axis)}")
        self._feature_dim_axis = feature_dim_axis
        if feature_dim:
            self.feature_dim = feature_dim

        if kwargs:
            self._handle_extra_kwargs(**kwargs)
        if raw_tensor is not None:
            self.raw_tensor = raw_tensor  # assignment via property, to have extra checks

    @property
    def dims(self) -> Tuple[Dim, ...]:
        """
        :return: dim tags
        """
        return self._dims

    @property
    def dims_set(self) -> Set[Dim]:
        """
        :return: set of dim tags. in all high-level code, the order of dims is irrelevant.
          The order must not play a role
          (RETURNN principles: https://github.com/rwth-i6/returnn/wiki/RETURNN-principles).
          Note that we do not include any implicit dims here.
          Also see :func:`verify_out_shape` and https://github.com/rwth-i6/returnn/issues/1153.
        """
        return set(self._dims)

    # This is potentially replaced by native implementation (Tensor._raw_tensor.__get__)
    @property
    def raw_tensor(self) -> Optional[RawTensorType]:
        """
        :return: raw tensor
        """
        return self._raw_tensor

    # This is potentially replaced by native implementation (_native tensor_raw_tensor_setter).
    @raw_tensor.setter
    def raw_tensor(self, value: Optional[RawTensorType]):
        """
        :param value:
        """
        if value is not None:
            # Small part from sanity _heck.
            # noinspection PyProtectedMember,PyShadowingNames
            import returnn.frontend._backend as _backend_api

            backend = _backend_api.get_backend_by_raw_tensor_type(type(value))
            raw_shape = backend.get_known_shape_raw(value)
            assert len(raw_shape) == len(self._dims), f"Mismatching shape ndim: Raw tensor {raw_shape} vs Tensor {self}"
            for i, dim in enumerate(self._dims):
                if dim.dimension is None:
                    continue  # we allow anything in the raw_tensor dim
                if raw_shape[i] != dim.dimension:
                    raise Exception(
                        f"Mismatching shape: Raw tensor {raw_shape} vs Tensor {self};\n"
                        + backend.format_graph_output(value, max_depth=3)
                    )
            if not backend.executing_eagerly():
                backend.set_known_shape_raw(value, self.batch_shape)
            assert backend.get_dtype_name_raw(value) == self.dtype, (
                f"{self} dtype {self.dtype} does not match raw tensor dtype {backend.get_dtype_name_raw(value)}"
            )
        self._raw_tensor = value

    @property
    def feature_dim(self) -> Optional[Dim]:
        """
        :return: self.dims[self.feature_dim_axis] or None.
            See https://github.com/rwth-i6/returnn/issues/1273 for some discussion.
        """
        # first fast paths
        if self._feature_dim_axis is None:
            return None
        if isinstance(self._feature_dim_axis, int):
            return self._dims[self._feature_dim_axis]
        if self.feature_dim_axis is None:
            return None
        return self._dims[self.feature_dim_axis]

    @feature_dim.setter
    def feature_dim(self, value: Optional[Dim]):
        """
        :param value:
        """
        if value is None:
            self._feature_dim_axis = None
            return
        assert not self.sparse_dim, "cannot have feature_dim_axis when sparse"
        self._feature_dim_axis = self.get_axis_from_description(value, allow_int=False)

    @property
    def device(self) -> Optional[str]:
        """
        :return: device
        """
        if self.raw_tensor is None:
            return None
        return self._raw_backend.get_device(self)



"""
Control flow context logic
"""

from __future__ import annotations
from typing import Optional
from contextlib import contextmanager
from returnn.tensor import ControlFlowContext


__all__ = ["control_flow_ctx", "get_current_control_flow_ctx"]


_ctx: Optional[ControlFlowContext] = None


@contextmanager
def control_flow_ctx(ctx: Optional[ControlFlowContext] = None):
    """
    Activates the given control flow context.
    """
    global _ctx
    prev_ctx = _ctx
    try:
        _ctx = ctx
        yield ctx
    finally:
        _ctx = prev_ctx


def get_current_control_flow_ctx() -> Optional[ControlFlowContext]:
    """
    :return: current control flow context
    """
    return _ctx



"""
Provides :class:`CachedDataset2`.
"""

from __future__ import annotations
import numpy
from typing import Optional, List
from threading import Condition
from .basic import Dataset, DatasetSeq


class CachedDataset2(Dataset):
    """
    Somewhat like CachedDataset, but different.
    Simpler in some sense. And more generic. Caching might be worse.

    If you derive from this class:
    - you must override `_collect_single_seq`
    - you must set `num_inputs` (dense-dim of "data" key) and `num_outputs` (dict key -> dim, ndim-1)
    - you should set `labels`
    - handle seq ordering by overriding `init_seq_order`
    - you can set `_estimated_num_seqs`
    - you can set `_num_seqs` or `_num_timesteps` if you know them in advance
    """

    def __init__(self, **kwargs):
        super(CachedDataset2, self).__init__(**kwargs)
        self._num_timesteps = None
        self.epoch = None
        self.reached_final_seq = False
        self.added_data: List[DatasetSeq] = []
        self.expected_load_seq_start = 0
        self._num_timesteps_accumulated = 0

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int|None epoch:
        :param list[str]|None seq_list: List of sequence tags, to set a predefined order.
        :param list[int]|None seq_order: List of corpus sequence indices, to set a predefined order. Only possible
          if the dataset has such indices (see self.have_corpus_seq_idx()).
        :rtype: bool
        :returns whether the order changed (True is always safe to return)

        This is called when we start a new epoch, or at initialization.
        Call this when you reset the seq list.
        """
        super(CachedDataset2, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        self.expected_load_seq_start = 0
        self.reached_final_seq = False
        self.added_data = []
        self._num_timesteps_accumulated = 0
        self._num_seqs = None
        self.epoch = epoch
        return True

    def _cleanup_old_seqs(self, seq_idx_end):
        """
        :param int seq_idx_end:
        """
        i = 0
        while i < len(self.added_data):
            if self.added_data[i].seq_idx >= seq_idx_end:
                break
            i += 1
        del self.added_data[:i]

    def _get_seq(self, seq_idx):
        """
        :param int seq_idx:
        :rtype: DatasetSeq|None
        """
        for data in self.added_data:
            if data.seq_idx == seq_idx:
                return data
        return None

    def is_cached(self, start, end):
        """
        :param int start:
        :param int end:
        :rtype: bool
        """
        # Always False, to force that we call self._load_seqs().
        # This is important for our buffer management.
        return False

    @property
    def num_seqs(self):
        """
        :rtype: int
        """
        if self._num_seqs is not None:
            return self._num_seqs
        if self.epoch is None:
            return 0
        raise NotImplementedError

    def _load_seqs(self, start, end):
        """
        :param int start: inclusive seq idx start
        :param int end: exclusive seq idx end. can be more than num_seqs
        If end > num_seqs, will not load them.
        """
        # We expect that start increase monotonic on each call
        # for not-yet-loaded data.
        # This will already be called with _load_seqs_superset indices.
        assert start >= self.expected_load_seq_start
        if start > self.expected_load_seq_start:
            # Cleanup old data.
            self._cleanup_old_seqs(start)
            self.expected_load_seq_start = start
        if self.added_data:
            start = max(self.added_data[-1].seq_idx + 1, start)
        seqs = [self._collect_single_seq(seq_idx=seq_idx) for seq_idx in range(start, end)]
        seqs = list(filter(None, seqs))  # We might not know the num seqs in advance.
        self._num_timesteps_accumulated += sum([seq.num_frames for seq in seqs])
        self.added_data += seqs

    def is_less_than_num_seqs(self, n):
        """
        :param int n:
        :rtype: int
        """
        if n < self.expected_load_seq_start:
            return True
        if self.epoch is None:
            return False
        # noinspection PyBroadException
        try:
            return super(CachedDataset2, self).is_less_than_num_seqs(n)
        except Exception:  # can fail, e.g. if self.num_seqs is not defined
            assert n >= self.expected_load_seq_start
            self._load_seqs(self.expected_load_seq_start, n + 1)
            if self._get_seq(n) is not None:
                return True
            # We reached the end.
            assert self.added_data, "Not a single seq was loaded?"
            self._num_seqs = self.added_data[-1].seq_idx + 1
            assert n >= self._num_seqs
            self.reached_final_seq = True
            return False

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        """
        :param seq_idx:
        :returns DatasetSeq or None if seq_idx >= num_seqs.
        """
        raise NotImplementedError

    def get_num_timesteps(self):
        """
        :rtype: int
        """
        if self._num_timesteps is not None:
            return self._num_timesteps
        else:
            assert self.reached_final_seq
            return self._num_timesteps_accumulated

    def _load_something(self):
        if self.added_data:
            return
        self.load_seqs(self.expected_load_seq_start, self.expected_load_seq_start + 1)
        assert self.added_data, f"{self}: _load_something: could not load anything"

    def get_seq_length(self, sorted_seq_idx):
        """
        :type sorted_seq_idx: int
        :rtype: returnn.util.NumbersDict
        """
        # get_seq_length() can be called before the seq is loaded via load_seqs().
        # Thus, we just call load_seqs() ourselves here.
        assert sorted_seq_idx >= self.expected_load_seq_start
        self.load_seqs(self.expected_load_seq_start, sorted_seq_idx + 1)
        return self._get_seq(sorted_seq_idx).num_frames

    def get_data(self, seq_idx: int, key: str) -> numpy.ndarray:
        """
        :param seq_idx:
        :param key:
        """
        return self._get_seq(seq_idx).features[key]

    def get_input_data(self, seq_idx: int) -> numpy.ndarray:
        """
        :param seq_idx:
        """
        return self.get_data(seq_idx, "data")

    def get_targets(self, target: str, seq_idx: int) -> numpy.ndarray:
        """
        :param target:
        :param seq_idx:
        """
        return self.get_data(seq_idx, target)

    def get_tag(self, sorted_seq_idx: int) -> str:
        """
        :param sorted_seq_idx:
        """
        # get_tag() can be called before the seq is loaded via load_seqs().
        # Thus, we just call load_seqs() ourselves here.
        self.load_seqs(self.expected_load_seq_start, sorted_seq_idx + 1)
        return self._get_seq(sorted_seq_idx).seq_tag

    def get_data_keys(self):
        """
        :rtype: list[str]
        """
        self._load_something()
        return sorted(self.added_data[0].get_data_keys())

    def get_target_list(self):
        """
        Target data keys are usually not available during inference.
        Overwrite this if your dataset is more custom.
        """
        keys = list(self.get_data_keys())
        if "data" in keys:
            keys.remove("data")
        return keys

    def get_complete_frac(self, sorted_seq_idx, **kwargs):
        """
        :return: fractional completion value for the given sorted_seq_idx
        """
        seq = self._get_seq(sorted_seq_idx)
        if seq is not None and seq.complete_frac is not None:
            return seq.complete_frac
        return super().get_complete_frac(sorted_seq_idx, **kwargs)

    def is_data_sparse(self, key):
        """
        :param str key: e.g. "data" or "classes"
        :rtype: bool
        """
        if key in self.num_outputs:
            return self.num_outputs[key][1] == 1
        self._load_something()
        return len(self.added_data[0].features[key].shape) == 1

    def get_data_dim(self, key):
        """
        :param str key: e.g. "data" or "classes"
        :rtype: int
        :return: number of classes, no matter if sparse or not
        """
        if key in self.num_outputs:
            d = self.num_outputs[key][0]
            if self.added_data and not self.is_data_sparse(key):
                assert self.added_data[0].get_data(key).shape[1] == d
            return d
        self._load_something()
        if len(self.added_data[0].get_data(key).shape) == 1:
            return super(CachedDataset2, self).get_data_dim(key)  # unknown
        assert len(self.added_data[0].get_data(key).shape) == 2
        return self.added_data[0].get_data(key).shape[1]

    def get_data_dtype(self, key):
        """
        :param str key:
        :rtype: str
        """
        self._load_something()
        return str(self.added_data[0].get_data(key).dtype)


class SingleStreamPipeDataset(CachedDataset2):
    """
    Producer: Gets data from somewhere / an external source, running in some thread.
    Consumer: The thread / code which calls load_seqs and get_data here.
    """

    def __init__(self, dim, ndim, sparse=False, dtype="float32"):
        """
        :param int dim:
        :param int ndim:
        :param bool sparse:
        :param str dtype:
        """
        super(SingleStreamPipeDataset, self).__init__()
        self.num_inputs = dim
        self.num_outputs = {"data": [dim, ndim]}
        self.sparse = sparse
        self.dtype = dtype
        self.condition = Condition()
        self.producer_seq_idx = 0
        self.producer_data = []
        self.producer_finished = False

    def is_data_sparse(self, key):
        """
        :param str key:
        :rtype: bool
        """
        return self.sparse

    def get_data_dtype(self, key):
        """
        :param str key:
        :rtype: str
        """
        return self.dtype

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :param int epoch:
        :param list[str]|None seq_list:
        :param list[int]|None seq_order:
        :rtype: bool
        """
        assert not seq_list and not seq_order
        super(SingleStreamPipeDataset, self).init_seq_order(epoch=epoch)
        with self.condition:
            self.producer_seq_idx = 0
            self.producer_data.clear()
            self.producer_finished = False
        return True

    def producer_add_data(self, data, seq_tag=None):
        """
        :param numpy.ndarray data:
        :param str|None seq_tag:
        """
        with self.condition:
            if seq_tag is None:
                seq_tag = "seq-%i" % self.producer_seq_idx
            seq = DatasetSeq(features=data, seq_idx=self.producer_seq_idx, seq_tag=seq_tag)
            self.producer_seq_idx += 1
            self.producer_data.append(seq)
            self.condition.notify()

    def producer_set_finished(self):
        """
        Mark finished.
        """
        with self.condition:
            self.producer_finished = True
            self.condition.notify()

    def _collect_single_seq(self, seq_idx):
        """
        :type seq_idx: int
        :rtype: DatasetSeq | None
        :returns DatasetSeq or None if seq_idx >= num_seqs.
        """
        with self.condition:
            while True:
                if self.producer_data:
                    seq = self.producer_data.pop(0)
                    assert isinstance(seq, DatasetSeq)
                    assert seq.seq_idx == seq_idx
                    return seq
                if self.producer_finished:
                    return None
                self.condition.wait()
